# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-06-29 03:00
from __future__ import unicode_literals

from django.db import migrations, models
import resolwe.flow.models.fields
import versionfield


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="TestModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
                ("slug", resolwe.flow.models.fields.ResolweSlugField()),
                ("version", versionfield.VersionField(default="0.0.0")),
            ],
        ),
    ]
